(require 'cl)
(require 'cleite-vars)
(require 'cleite-feeds)

(defvar cleite:items:buffer-name "*cleite feed items*")

(defvar cleite:items:mode-map
  (let ((map (make-sparse-keymap)))
	(set-keymap-parent map cleite:global-map)
	
	(define-key map "\r" 'cleite:items:select)
    (define-key map "U"  'cleite:items:copy-url)
	(define-key map "s"  'cleite:items:refresh)
    
	(define-key map "N"  'cleite:items:go-to-next-unread)
	(define-key map "P"  'cleite:items:go-to-prev-unread)
	(define-key map "n"  'cleite:items:go-to-next)
	(define-key map "p"  'cleite:items:go-to-prev)
    
    (define-key map "u"  'cleite:items:mark-as-unread)
    (define-key map "r"  'cleite:items:mark-as-read)
    (define-key map "mu" 'cleite:items:mark-region-as-unread)
    (define-key map "mr" 'cleite:items:mark-region-as-read)

    (define-key map "q"  'cleite:items:quit)
	map)
  "Keymap for cleite:items:mode")


(defvar cleite:items:mode-menu-spec
  '("Feed Items"
	["Select article"                cleite:items:select               ]
    ["Copy URL..."                   cleite:items:copy-url             ]
	["Refresh article list"          cleite:items:refresh              ]
	"----"
	["Go to next unread article"     cleite:items:go-to-next-unread    ]
	["Go to previous unread article" cleite:items:go-to-prev-unread    ]
	["Go to next article"            cleite:items:go-to-next           ]
	["Go to previous article"        cleite:items:go-to-prev           ]
    "----"
    ["Mark as unread"                cleite:items:mark-as-unread       ]
    ["Mark as read"                  cleite:items:mark-as-read         ]
    ["Mark region as unread"         cleite:items:mark-region-as-unread]
    ["Mark region as read"           cleite:items:mark-region-as-read  ]
	"----"
	["Close article list"            cleite:items:quit                 ]))


(defvar cleite:items:mode-hook nil
  "Hook for cleite:items:mode")

(defvar cleite:items:feed-uri nil)
(defvar cleite:items:feed     nil)
(defvar cleite:items:list     nil)


(easy-menu-define cleite:items:mode-menu
  cleite:items:mode-map
  "Menu used in cleite:items:mode"
  cleite:items:mode-menu-spec)


(defun* cleite:items:current-feed-uri (&key clear)
  (when clear
	(setq cleite:items:feed-uri nil))
  cleite:items:feed-uri)


(defun cleite:items:unread-item-p (item)
  (eq 'false
	  (cdr (assq 'read-before item))))


(defun cleite:items:show-status (status)
  (cleite:items:on-items-buffer
	
	(if (or (null status)
			(= 0 (length status)))
		(setq mode-name "Cleite Feed Items")
	  (setq mode-name (format "Cleite Feed Items:%s" status)))

	(force-mode-line-update)
    (sit-for 0)))


;;;###autoload
(defun cleite:items:mode ()
  "Major mode for Cleite feed items view
\\{cleite:items:mode-map}"
  (interactive)
  (kill-all-local-variables)
  (setq major-mode 'cleite:items:mode)
  (cleite:items:show-status nil)
  (use-local-map cleite:items:mode-map)
  (easy-menu-add cleite:items:mode-menu)
  (run-hooks 'cleite:items:mode-hook))


;;;###autoload
(defun* cleite:items (feed-uri &key refresh)
  "items バッファがあれば呼び出し、無ければ作る。

REFRESH が t の場合は、バッファ全體を常に再構築する。
項目 URI 文字列の場合は、該當する行を常に再構築する。"

  ;; feeds バッファのみを表示する。
  (cleite:feeds)

  ;; feeds バッファを表示してゐるウインドウを左右に分割する。
  (split-window-horizontally cleite:feeds:window-width)
  (other-window 1)

  ;; items バッファがあれば呼び出し、無ければ作る。
  (if (get-buffer cleite:items:buffer-name)
	  (progn
		(switch-to-buffer cleite:items:buffer-name)
		;; もし指定された url が現在選ばれてゐるものと違ふのなら、ここ
		;; で常に refresh する。
		(cond ((equal feed-uri cleite:items:feed-uri)
			   (cond ((eq t refresh)
					  (cleite:items:refresh))

					 ((stringp refresh)
					  (cleite:items:rebuild-line refresh))))

			  (t
			   (goto-char (point-min))
			   (setq cleite:items:feed-uri feed-uri)
			   (cleite:items:refresh))))

	(switch-to-buffer (get-buffer-create cleite:items:buffer-name))
	(cleite:items:mode)
	
	(setq cleite:items:feed-uri feed-uri)
	(cleite:items:refresh)))


(defun cleite:items:refresh ()
  (interactive)
  (when cleite:items:feed-uri
    
    (cleite:items:show-status "Loading feed items...")

    (setq cleite:items:feed (cleite:call 'feed:get-info
                                         `((feed-uri . ,cleite:items:feed-uri))))
  
    (setq cleite:items:list (cleite:call 'feed-item:list
                                         `((feed-uri . ,cleite:items:feed-uri))))
    (cleite:items:rebuild-buffer)
  
    (cleite:items:show-status nil)))


(defun cleite:items:sort-comparator (a b)
  (block exit
	(labels ((read-p (item)
			     (eq 'true (cdr (assq 'read-before item))))

			 (date-of (item)
				 (cdr (assq 'date item)))

			 (title-of (item)
				 (cdr (assq 'title item)))

			 (negate-if-descending (bool order)
				 (if (or (eq order 'descending)
						 (eq order 'read-first))
					 (not bool)
				   bool)))

	  (dolist (criteria cleite:items:sort-criteria)
		(let ((type  (car criteria))
			  (order (cdr criteria)))

		  (cond ((eq type 'read-flag)
				 (if (not (eq (read-p a) (read-p b)))
					 (return-from exit (negate-if-descending
										   (and (not (read-p a)) (read-p b)) order))))

				((eq type 'date)
				 (if (not (= (date-of a) (date-of b)))
					 (return-from exit (negate-if-descending
										   (< (date-of a) (date-of b)) order))))

				((eq type 'title)
				 (if (not (equal (title-of a) (title-of b)))
					 (return-from exit (negate-if-descending
										   (> 0 (compare-strings (title-of a) 0 nil
																 (title-of b) 0 nil)) order))))

				(t
				 ;; 不正な type
				 (error "invalid criterion type: %s" type)))))

	  ;; 結局決まらなかったので兩者は等しい事にする。
	  nil)))


(defun cleite:items:sort ()
  (setq cleite:items:list
		(sort cleite:items:list #'cleite:items:sort-comparator)))

  
(defun cleite:items:rebuild-buffer ()
  (cleite:items:on-items-buffer
	(let ((saved-point (point)))
	  
	  (setq buffer-read-only nil)
	  (erase-buffer)

	  ;; モード行
	  (setq mode-line-buffer-identification
			(cdr (assq 'title cleite:items:feed)))

	  (cleite:items:sort)

	  (dolist (item cleite:items:list)

		(unless (= (point) (point-min))
		  (insert "\n"))

		(cleite:items:insert-line item))

	  (setq buffer-read-only t)
	  (goto-char saved-point)
	  
	  (when (and cleite:items:jump-to-unread
                 (get-buffer-window cleite:items:buffer-name)
                 (not (cleite:article:current-item-uri)))
		(cleite:items:find-next-item :unread-only t :no-skip t)))))


(defun cleite:items:rebuild-line (item-uri)
  (cleite:items:on-items-window
	(setq buffer-read-only nil)
	
	(block nil
	  (save-excursion
		(let* ((line-num    (or (cleite:items:item-position item-uri)
								;; 無ければ逃げる
								(return nil)))
			   (item        (nth line-num cleite:items:list))

			   (begin-point (progn
							  (goto-char (point-min))
							  (forward-line line-num)
							  (point)))

			   (end-point   (progn
							  (end-of-line)
							  (point))))

		  ;; この行端から行末までを再構築する。
		  (delete-region begin-point end-point)
		  (cleite:items:insert-line item))))

	(setq buffer-read-only t)))


(defun cleite:items:insert-line (item)
  (let* ((title        (cdr (assq 'title    item)))
		 (item-uri     (cdr (assq 'item-uri item)))
		 
		 (read-before  (if (eq 'true (cdr (assq 'read-before item)))
						  t nil))

		 (default-face (if read-before
						   'cleite:face:read-item
						 'cleite:face:unread-item))
		 
		 (title-face   (if (equal item-uri (cleite:article:current-item-uri))
						   'cleite:face:selected-item
						 default-face))
					  
		 (prev         (point)))

    ;; タイトルに改行コードが入ってゐると不都合なので、改行コードは削除
    ;; する。
    (setq title (replace-regexp-in-string "\n" "" title))

    ;; タイトルが空文字列の場合は、何でも良いので空以外にしないと撰擇出
    ;; 來なくなってしまふ。
    (when (equal "" title)
      (setq title "(empty title)"))
    
	(insert (cleite:faced-string title title-face))

	;; 今バッファに插入した部分についてテキスト屬性を設定する。
	(put-text-property prev (point) 'mouse-face  'highlight)
	(put-text-property prev (point) 'cleite:item item      )))


(defun cleite:items:item-position (item-uri &optional list)
  (position-if (lambda (item)
				 (let ((uri (cdr (assq 'item-uri item))))
				   (equal uri item-uri)))
			   (or list cleite:items:list)))


(defun cleite:items:set-read (item-uri flag)
  (block nil
	;; set-read メソッドを呼び、get メソッドで更新後のデータを得る。但
	;; し既に read フラグが立ってゐるなら何もしない。
	(let* ((feed-uri cleite:items:feed-uri)
		   (pos      (cleite:items:item-position item-uri))
		   (item     (nth pos cleite:items:list))
		   (is-read  (eq 'true (cdr (assq 'read-flag item))))
		   (flag-sym (if flag
						 'true 'false)))

	  (if is-read
		  (return nil))
	
	  (cleite:call 'feed-item:set-read
				   `((feed-uri    . ,feed-uri)
					 (item-uri    . ,item-uri)
					 (read-before . ,flag-sym)))

      ;; 新たな item を feed-item:get を使って取った方が確實だが、回線
      ;; に遲延がある時は大きなストレスになる。
      (let ((fresh-item (cons `(read-before . ,flag-sym)
                              (assq-delete-all 'read-before item))))
		(replace cleite:items:list
				 (list fresh-item)
				 :start1 pos :end1 (1+ pos)))

	  ;; feeds 側にも更新させる。フィードから未讀項目が消えた時は、フィー
	  ;; ドリスト全體をソートし直す必要がある。
	  (cleite:feeds:refresh :feed-uri feed-uri))))


(defun cleite:items:select ()
  (interactive)
  (cleite:items:with-pointed-item item

	;; 撰擇されてゐる項目があればバッファを更新
	(let ((item-uri (cleite:article:current-item-uri)))
	  (when item-uri
		(cleite:article:current-item-uri :clear t)
		(cleite:items:rebuild-line item-uri)))

	;; 既讀にして、article ウインドウを開いてからバッファを更新
	(let ((feed-uri (cdr (assq 'feed-uri item)))
		  (item-uri (cdr (assq 'item-uri item))))

	  (cleite:items:set-read item-uri t)
	  (cleite:article feed-uri item-uri)
	  (cleite:items:rebuild-line item-uri))))


(defun cleite:items:mark-as-unread ()
  (interactive)
  (cleite:items:with-pointed-item item
    (let ((item-uri (cdr (assq 'item-uri item))))
      
      (cleite:items:set-read item-uri nil)
      (cleite:items:rebuild-line item-uri))))


(defun cleite:items:mark-as-read ()
  (interactive)
  (cleite:items:with-pointed-item item
    (let ((item-uri (cdr (assq 'item-uri item))))
      
      (cleite:items:set-read item-uri t)
      (cleite:items:rebuild-line item-uri))))


(defun cleite:items:mark-region-as-unread ()
  (interactive)
  (cleite:items:set-read-flag-in-region nil))


(defun cleite:items:mark-region-as-read ()
  (interactive)
  (cleite:items:set-read-flag-in-region t))


(defun cleite:items:set-read-flag-in-region (flag)
  (block nil

    ;; マークが無ければ終了
    (unless (mark)
      (return nil))

  	 (beginning-of-line)
     (let ((begin-item (get-text-property (mark)  'cleite:item))
           (end-item   (get-text-property (point) 'cleite:item)))

       (unless (and begin-item end-item)
         ;; マーク位置とカーソル位置が共に項目上に無ければ終了
         (return nil))

       (let ((begin-pos (cleite:items:item-position
                         (cdr (assq 'item-uri begin-item))))
             
             (end-pos   (cleite:items:item-position
                         (cdr (assq 'item-uri end-item)))))

         (loop for i
               from (if (<= begin-pos end-pos)
                        begin-pos end-pos)
               to   (if (>= end-pos begin-pos)
                        end-pos begin-pos)
               do
               (let* ((item     (nth i cleite:items:list))
                      (item-uri (cdr (assq 'item-uri item))))

                 (cleite:items:set-read item-uri flag)
                 (cleite:items:rebuild-line item-uri)))))))

(defun cleite:items:copy-url ()
  (interactive)
  (let* ((feed     cleite:items:feed)
         (feed-uri (cdr (assq 'feed-uri feed)))
         (site-uri (cdr (assq 'site-uri feed)))
         (title    (cdr (assq 'title    feed)))
         (choice   (cleite:read-char-with-retry
                    (format "Copy site [u]rl / Copy [t]itle / Copy [b]oth / Copy [f]eed url: `%s': "
                            title)
                    '(?u ?t ?b ?f))))

    (let ((str (cond ((eq choice ?u)
                      site-uri)

                     ((eq choice ?t)
                      title)

                     ((eq choice ?b)
                      (format "%s\n%s" title site-uri))

                     ((eq choice ?f)
                      feed-uri))))
      (kill-new str)
      (message "Copied: %s" str))))


(defun cleite:items:quit (&optional kill)
  (interactive "P")

  (block nil
	(let ((buf (get-buffer cleite:items:buffer-name)))

	  (if (null buf)
		  (return nil))

	  (delete-windows-on buf)

	  (if kill
		  ;; ただバッファを殺すだけ
		  (kill-buffer buf)
		;; feeds バッファを更新
		(let ((feed-uri cleite:items:feed-uri))
		  (setq cleite:items:feed-uri nil)
		  (cleite:feeds :refresh feed-uri)))))

  (setq cleite:items:feed-uri nil)
  (setq cleite:items:feed     nil)
  (setq cleite:items:list     nil))


(defmacro cleite:items:with-pointed-item (item-sym &rest rest)
  (declare (indent 1))
  `(block nil
	 (beginning-of-line)
	 (let ((,item-sym (get-text-property (point) 'cleite:item)))
	   
	   (if (null ,item-sym)
		   ;; カーソルが項目上でない場所にあるので何もせずに終了
		   (return nil))

	   ,@rest)))


(defmacro cleite:items:on-items-window (&rest rest)
  "items バッファのウインドウが撰擇された状態で REST を實行する。

ウインドウの撰擇までは必要無い時は `cleite:items:on-items-buffer'
を使ふ事。"
  (declare (indent 0))
  `(with-selected-window
	   (get-buffer-window cleite:items:buffer-name)
     (cleite:items:on-items-buffer
       ,@rest)))


(defmacro cleite:items:on-items-buffer (&rest rest)
  "items バッファが撰擇された状態で REST を實行する。

ウインドウの撰擇が必要な時は `cleite:items:on-items-window' を使ふ。"
  (declare (indent 0))
  `(with-current-buffer cleite:items:buffer-name
     ,@rest))


(defun cleite:items:go-to-next ()
  (interactive)
  (if (cleite:items:find-next-item-among-feeds)
	  t
	(message "This is the last item of the last feed.")
	nil))


(defun cleite:items:go-to-prev ()
  (interactive)
  (if (cleite:items:find-next-item-among-feeds :reverse t)
	  t
	(message "This is the first item of the first feed.")
	nil))


(defun cleite:items:go-to-next-unread ()
  (interactive)
  (if (cleite:items:find-next-item-among-feeds :unread-only t)
	  t
	(message "This is the last unread item of the last unread feed.")
	nil))


(defun cleite:items:go-to-prev-unread ()
  (interactive)
  (if (cleite:items:find-next-item-among-feeds :unread-only t :reverse t)
	  t
	(message "This is the first unread item of the first unread feed.")
	nil))


(defun* cleite:items:find-next-item-among-feeds (&key unread-only reverse)
  (flet ((go-to-next-feed ()
           (block nil
             (while t
               (if (cleite:feeds:find-next-feed :unread-only unread-only
                                                :reverse     reverse    )
                   ;; フィード移動成功
                   (progn
                     ;; 先づこのフィードを撰擇する。
                     (cleite:feeds:on-feeds-window
                       (cleite:feeds:select)
                       (cond (reverse
                              (goto-char (point-max))
                              (beginning-of-line))
                             (t
                              (goto-char (point-min)))))
                     ;; 項目を檢索
                     (when (cleite:items:find-next-item :unread-only unread-only
                                                        :reverse     reverse
                                                        :no-skip     t          )
                       ;; このフィード内に該當する項目が存在した。
                       (return t)))
                 ;; もうフィードが無い
                 (return nil))))))
	(if (cleite:items:find-next-item :unread-only unread-only
									 :reverse     reverse    )
		;; 項目移動成功
		t
	  ;; このフィードにはもう該當する項目が無いので、次のフィードへ行け
	  ;; るなら行く。
	  (go-to-next-feed))))


(defun* cleite:items:find-next-item (&key unread-only reverse no-skip)
  (cleite:items:on-items-window
	(cleite:items:with-pointed-item
		item
	  ;; 項目上にカーソルがある
	  (let* ((item-uri (cdr (assq 'item-uri item)))
		   
			 (list     (if reverse
						   (reverse cleite:items:list)
						 cleite:items:list))
		   
			 (pos      (cleite:items:item-position item-uri list))
		   
			 (rest     (nthcdr (if no-skip
								   pos
								 (1+ pos))
							   list))
		   
			 (slip     (if reverse
						   -1
						 1)))

		;; no-skip が僞である時は、既に rest の要素數が一つ減らされてゐる。
		(unless no-skip
		  (forward-line slip)
		  (beginning-of-line))

        ;; 末尾再歸で書きたい所だが、Elisp でそれは出來ない。
        (block nil
          ;; 檢索對象がもう無ければ失敗。
          (dolist (item rest nil)
            (if unread-only
                ;; 未讀なら成功。未讀でなければ次へ。
                (if (cleite:items:unread-item-p item)
                    (return t)
                  (forward-line slip)
                  (beginning-of-line))
              
              ;; 常に成功
              (return t))))))))


(provide 'cleite-items)